from dataclasses import make_dataclass, field
from typing import Dict, Tuple, Any, Optional, List, Iterator, Type

from pydantic import BaseModel, create_model, Field

from cass_api.assistant.ddl_tool import DDLModel, Column
from datastore.providers.cassandra_util import get_pydantic_type, python_to_cassandra
from datastore.providers.simple_cassandra_datastore import CassandraDataStore


class LoginPayload(BaseModel):
    db_id: str

datastores = {}

def get_datastore_from_cache(token) -> CassandraDataStore:
    global datastores
    if token in datastores:
        return datastores[token]
    raise Exception(detail="Must login to a database first")

def db_login(payload: LoginPayload, token: str):
    global datastores
    datastore = datastores.get(token)
    if datastore is None:
        datastore = CassandraDataStore()
    if payload.db_id == "":
        return Exception(detail='{"msg": "db_id is required."}')
    datastore.setupSession(token, payload.db_id)
    datastores[token] = datastore


class Table:
    def __init__(self, db, table_name):
        self.db = db
        self.table_name = table_name
        self.keyspace = db.keyspace
        self.columns = db.table_columns(table_name)

        model_name = ''.join(word.capitalize() for word in self.keyspace.split('_'))
        model_name += ''.join(word.capitalize() for word in self.table_name.split('_'))
        model_name += "Model"

        columns = self.db.client.get_columns(self.keyspace, self.table_name)
        model_fields: Dict[str, Tuple[Any, Any]] = {}
        dataclass_fields: List[Tuple[str, Any]] = []  # Fields for dataclass
        for col in columns:
            if col['kind'] != 'partition_key' and col['kind'] != 'clustering':
                model_fields[col["column_name"]] = (Optional[get_pydantic_type(col["type"])], None)
                dataclass_fields.append((col["column_name"], get_pydantic_type(col["type"])))
            else:
                model_fields[col["column_name"]] = (get_pydantic_type(col["type"]), ...)
                dataclass_fields.append((col["column_name"], get_pydantic_type(col["type"])))

        ResponseModel = create_model(model_name, **model_fields)
        Dataclass = make_dataclass(model_name, dataclass_fields)
        self._model = ResponseModel
        self._dataclass = Dataclass

    def exists(self) -> bool:
        tables = self.db.client.get_tables(self.keyspace)
        return self.table_name in tables

    def __repr__(self) -> str:
        return "<Table {}{}>".format(
            self.table_name,
            (
                " (does not exist yet)"
                if not self.exists()
                else " ({})".format(", ".join(c for c in self.columns))
            ),
        )

    def pydantic_model(self):
        return self._model


    def dataclass(self):
        return self._dataclass


    def all(self) -> List[BaseModel]:
        rows = self.db.client.select_all_from_table(self.keyspace, self.table_name)
        return [self._model(**row) for row in rows]

    def create(
            self,
            partition_key: List[str] = None,
            clustering_columns: List[str] = None,
            columns: Dict[str, Any] = None,
            **kwargs
    ):
        if not columns:
            columns={}
        columns = {**columns, **kwargs}
        column_list = []
        if not partition_key:
            partition_key = []
        if isinstance(partition_key, str):
            partition_key = [partition_key]
        if not clustering_columns:
            clustering_columns = []
        if isinstance(clustering_columns, str):
            clustering_columns = [clustering_columns]
        for column_name, column_type in columns.items():
            column_list.append(Column(name=column_name, type=python_to_cassandra(column_type)))
        ddl_model = DDLModel(
            keyspace_name=self.db.keyspace,
            table_name=self.table_name,
            columns=column_list,
            partition_key=partition_key,
            clustering_columns=clustering_columns,
            thoughts=None
        )
        self.db.client.execute(ddl_model.to_string())
        self.columns = self.db.table_columns(self.table_name)

    @property
    def c(self):
        rows = self.db.client.get_columns(self.keyspace, self.table_name)
        return rows

    def insert(self, request_object: BaseModel = None, **kwargs):
        if request_object is None:
            request_object = self._model(**kwargs)
        self.db.client.upsert_table_from_dict(self.keyspace, self.table_name, request_object.dict())

    def update(self, request_object: BaseModel = None, **kwargs):
        self.insert(request_object, **kwargs)


class DynamicTables:
    def __init__(self, db, tables):
        self.db = db
        self._tables = tables

    def __getattr__(self, table_name) -> Table:
        for table in self._tables:
            if table.table_name == table_name:
                return table
        table = Table(self.db, table_name)
        return table

    def __dir__(self):
        table_names = []
        for table in self._tables:
            table_names.append(table.table_name)
        return self.db.client.get_tables(self.db.keyspace) + table_names

    def __iter__(self) -> Iterator[Table]:
        return iter(self._tables)



class Database:
    def __init__(self, token, dbid):
        login_payload = LoginPayload(db_id=dbid)
        db_login(login_payload, token)
        datastore = get_datastore_from_cache(token)
        self.client = datastore.client
        self.keyspace = "default_keyspace"
        self._tables = None

    def __del__(self):
        pass

    @property
    def t(self):
        if self._tables is None:
            rows = self.client.get_tables(self.keyspace)
            tables = []
            for row in rows:
                tables.append(row)
            self._tables = [Table(self, table) for table in tables]
        return DynamicTables(self, self._tables)

    def table_columns(self, table_name):
        rows = self.client.get_columns(self.keyspace, table_name)
        columns = []
        for row in rows:
            columns.append(row["column_name"])
        return columns




import subprocess
import os
import threading

from fastapi import HTTPException

from cass_api.ctags.util import generate_ctags

# Create a global lock for thread safety
lock = threading.Lock()

# Define paths
target_sdk_path = './fern_config/sdks/python/src/'
fern_config_path = './fern_config'

def run_command(command, cwd=None):
    """
    Run a shell command and handle errors.
    """
    try:
        if cwd:
            cwd = os.path.abspath(cwd)

        full_command = command
        if os.path.exists(os.path.expanduser("~/.nvm")):
            full_command = f"bash -c 'export NVM_DIR=\"$HOME/.nvm\" && source $NVM_DIR/nvm.sh && nvm use lts/iron && {command}'"
        result = subprocess.run(full_command, check=True, shell=True, cwd=cwd, capture_output=True, text=True)
        return result.stdout
    except subprocess.CalledProcessError as e:
        print(f"Error running command: {e}")
        print(f"Output: {e.output}")
        print(f"Error Output: {e.stderr}")

def generate_sdk(openapi_content, ctags_file_name):
    """
    Write the openapi.yaml content to the specified destination and generate the SDK.
    """
    openapi_destination = os.path.join(fern_config_path, 'fern/openapi/openapi.yaml')

    with lock:
        os.makedirs(os.path.dirname(openapi_destination), exist_ok=True)
        with open(openapi_destination, 'w') as file:
            file.write(openapi_content)

        try:
            # Run the fern generate commands
            print(run_command(f'fern generate --group python-sdk', cwd=fern_config_path))

            # Generate ctags
            generate_ctags(target_sdk_path, f"./tenant_ctags/{ctags_file_name}")

        except subprocess.CalledProcessError as e:
            print(f"Error generating SDK: {e}")
            HTTPException(status_code=500, detail="Error generating SDK")


from abc import ABC, abstractmethod


class DataStore(ABC):
    @abstractmethod
    def setupEndpoints(self, app):
        """
        Sets up the app endpoints given existing tables available in the datastore.
        """
        raise NotImplementedError


    def setupSession(self, token, dbid):
        """
        Sets up the datastore session given a token.
        """
        raise NotImplementedError

    def getSubApp(self):
        """
        Sets up the datastore session given a token.
        """
        raise NotImplementedError
